package com.icing.dialer

import android.util.Base64
import com.southernstorm.noise.protocol.CipherStatePair
import com.southernstorm.noise.protocol.HandshakeState
import com.southernstorm.noise.protocol.Noise
import java.util.Arrays
import javax.crypto.BadPaddingException
import javax.crypto.ShortBufferException
import java.security.NoSuchAlgorithmException

class NoiseHandlerXX(
    private val localKeyBase64: String // ED25519 private key (Base64-encoded)
) {
    private var handshakeState: HandshakeState? = null
    private var cipherStatePair: CipherStatePair? = null
    private var remotePublicKeyBase64: String? = null

    /**
     * Wipes sensitive data by filling the byte array with zeros.
     */
    private fun wipe(data: ByteArray?) {
        data?.let { Arrays.fill(it, 0.toByte()) }
    }

    /**
     * Initializes the Noise XX handshake.
     * @param isInitiator True if this is the initiator, false if responder.
     * @return The initial handshake message.
     * @throws IllegalArgumentException If local key is invalid.
     * @throws IllegalStateException If handshake fails to start.
     */
    fun initialize(isInitiator: Boolean): ByteArray {
        var localKey: ByteArray? = null
        try {
            val protocolName = "Noise_XX_25519_AESGCM_SHA256"
            handshakeState = HandshakeState(
                protocolName,
                if (isInitiator) HandshakeState.INITIATOR else HandshakeState.RESPONDER
            )

            // Set local private key
            localKey = Base64.decode(localKeyBase64, Base64.DEFAULT)
            if (localKey.size != 32) {
                throw IllegalArgumentException("Invalid local key size: ${localKey.size}")
            }
            handshakeState?.localKeyPair?.setPrivateKey(localKey, 0)
                ?: throw IllegalStateException("Local key pair not initialized")

            // Start handshake and write initial message
            handshakeState?.start() ?: throw IllegalStateException("Handshake state not initialized")
            val keyLength = handshakeState?.localKeyPair?.getPublicKeyLength() ?: 32
            val macLength = 16 // AES-GCM MAC length
            val messageBuffer = ByteArray(keyLength + macLength + 128) // Conservative sizing
            val payload = ByteArray(0) // Empty payload
            val writtenLength: Int = handshakeState?.writeMessage(
                messageBuffer, 0, payload, 0, payload.size
            ) ?: throw IllegalStateException("Failed to write handshake message")
            return messageBuffer.copyOf(writtenLength)
        } catch (e: NoSuchAlgorithmException) {
            throw IllegalStateException("Unsupported algorithm: ${e.message}", e)
        } catch (e: ShortBufferException) {
            throw IllegalStateException("Buffer too small for handshake message", e)
        } finally {
            wipe(localKey)
        }
    }

    /**
     * Processes a handshake message and returns the next message or null if complete.
     * @param message The received handshake message.
     * @return The next handshake message, null if handshake is complete, or the remote public key (Base64-encoded).
     * @throws IllegalStateException If handshake state is invalid.
     * @throws BadPaddingException If message decryption fails.
     */
    fun processHandshakeMessage(message: ByteArray): Pair<ByteArray?, String?> {
        var remotePublicKey: ByteArray? = null
        try {
            val handshake = handshakeState ?: throw IllegalStateException("Handshake not initialized")
            val keyLength = handshake.localKeyPair?.getPublicKeyLength() ?: 32
            val macLength = 16 // AES-GCM MAC length
            val messageBuffer = ByteArray(keyLength + macLength + 128) // Conservative sizing
            val payload = ByteArray(0) // Empty payload
            val writtenLength: Int = handshake.readMessage(
                message, 0, message.size, messageBuffer, 0
            )

            // Extract remote public key after reading message
            if (handshake.remotePublicKey.hasPublicKey()) {
                remotePublicKey = ByteArray(keyLength)
                handshake.remotePublicKey.getPublicKey(remotePublicKey, 0)
                remotePublicKeyBase64 = Base64.encodeToString(remotePublicKey, Base64.DEFAULT)
            }

            if (handshake.getAction() == HandshakeState.SPLIT) {
                cipherStatePair = handshake.split()
                return Pair(null, remotePublicKeyBase64) // Handshake complete
            }

            // Write next message
            val nextMessage = ByteArray(keyLength + macLength + 128)
            val nextWrittenLength: Int = handshake.writeMessage(
                nextMessage, 0, payload, 0, payload.size
            )
            return Pair(nextMessage.copyOf(nextWrittenLength), remotePublicKeyBase64)
        } catch (e: ShortBufferException) {
            throw IllegalStateException("Buffer too small for handshake message", e)
        } catch (e: BadPaddingException) {
            throw IllegalStateException("Invalid handshake message: ${e.message}", e)
        } finally {
            wipe(remotePublicKey)
        }
    }

    /**
     * Encrypts data using the sender's cipher state.
     * @param data The data to encrypt.
     * @return The encrypted data.
     * @throws IllegalStateException If handshake is not completed.
     */
    fun encryptData(data: ByteArray): ByteArray {
        val cipherState = cipherStatePair?.getSender()
            ?: throw IllegalStateException("Handshake not completed")
        try {
            val outputBuffer = ByteArray(data.size + cipherState.getMACLength()) // Account for AES-GCM MAC
            val length: Int = cipherState.encryptWithAd(null, data, 0, outputBuffer, 0, data.size)
            return outputBuffer.copyOf(length)
        } catch (e: ShortBufferException) {
            throw IllegalStateException("Buffer too small for encryption: ${e.message}", e)
        }
    }

    /**
     * Decrypts data using the receiver's cipher state.
     * @param data The encrypted data.
     * @return The decrypted data.
     * @throws IllegalStateException If handshake is not completed.
     * @throws BadPaddingException If decryption fails.
     */
    fun decryptData(data: ByteArray): ByteArray {
        val cipherState = cipherStatePair?.getReceiver()
            ?: throw IllegalStateException("Handshake not completed")
        try {
            val outputBuffer = ByteArray(data.size)
            val length: Int = cipherState.decryptWithAd(null, data, 0, outputBuffer, 0, data.size)
            return outputBuffer.copyOf(length)
        } catch (e: ShortBufferException) {
            throw IllegalStateException("Buffer too small for decryption: ${e.message}", e)
        } catch (e: BadPaddingException) {
            throw IllegalStateException("Invalid ciphertext: ${e.message}", e)
        }
    }

    /**
     * Returns the remote public key received during the handshake.
     * @return The Base64-encoded remote public key, or null if not yet received.
     */
    fun getRemotePublicKey(): String? {
        return remotePublicKeyBase64
    }

    /**
     * Cleans up sensitive cryptographic data.
     */
    fun destroy() {
        handshakeState?.destroy()
        cipherStatePair?.destroy()
        handshakeState = null
        cipherStatePair = null
        remotePublicKeyBase64 = null
    }
}